use crate::router::ReqResult;
use crate::state::State as AppState;
use crate::{partials::*, user::User};
use axum::extract::State;
use axum::response::Html;
use axum::Form;
use maud::html;
use redact::Secret;
use serde::Deserialize;
use std::convert::Infallible;

use axum::{
    routing::{get, post},
    Router,
};

use crate::{prelude::*, user};

pub fn router(state: AppState) -> Result<Router, Infallible> {
    Ok(Router::new()
        .route("/login", get(login))
        .route("/login", post(authenticate))
        .route("/register", get(register))
        .route("/register", post(create_user))
        .with_state(state))
}

async fn login() -> ReqResult<Html<String>> {
    let form = html! {
        (labelled_input("Username", html!{
            input class="input" type="text" name="username" autocomplete="username" required;
        }))
        (labelled_input("Password", html!{
            input class="input" type="password" name="password" autocomplete="current-password" required;
        }))
    };
    let subaction = html! {
        small class="mt-4" {
            "Need an account? "
            a href="/auth/register" {"Get one"}
            "."
        }
    };
    page("login", center_hero_form("Login", form, subaction))
}

async fn register() -> ReqResult<Html<String>> {
    let form = html! {
        (labelled_input("Username", html!{
            input class="input" type="text" name="username" required;
        }))
        (labelled_input("Password", html!{
            input class="input" type="password" name="password" autocomplete="new-password" required;
        }))
    };
    let subaction = html! {
        small class="mt-4" {
            "Already have an account? "
            a href="/auth/login" {"Login"}
            "."
        }
    };
    page("login", center_hero_form("Register", form, subaction))
}

#[derive(Deserialize, Debug)]
struct Creds {
    username: String,
    password: Secret<String>,
}

#[instrument(skip(state))]
async fn authenticate(
    State(state): State<AppState>,
    Form(creds): Form<Creds>,
) -> ReqResult<Html<String>> {
    let existing_user: Option<User> = state.db.get(User::tree(), &creds.username)?;

    if existing_user.is_none() {
        // timing/enumeration attacks or something
        return Err(user::Error::UsernameNotFound(Box::new(creds.username)).into());
    }

    let existing_user = existing_user.unwrap();
    if let Err(err) = existing_user.verify(creds.password.expose_secret()) {
        Ok(Html(
            html! {
                "failed to login: " ({err})
            }
            .into_string(),
        ))
    } else {
        Ok(Html(
            html! {
                "logged in"
            }
            .into_string(),
        ))
    }
}

#[instrument(skip(state))]
async fn create_user(
    State(state): State<AppState>,
    Form(creds): Form<Creds>,
) -> ReqResult<Html<String>> {
    let user = User::try_new(&creds.username, creds.password.expose_secret())?;
    let existing_user: Option<User> = state.db.get(User::tree(), &creds.username)?;

    // TODO: fail2ban?
    if existing_user.is_some() {
        // timing/enumeration attacks or something
        return Err(user::Error::UsernameExists(Box::new(creds.username)).into());
    }

    state
        .db
        .insert(User::tree(), &user.username, bincode::serialize(&user)?)?;

    Ok(Html(
        html! {
            ({&user.username}) " has been registered"
        }
        .into_string(),
    ))
}
